@@ -1,16 +1,20 @@
 // Copyright 2019-2022 ChainSafe Systems
 // SPDX-License-Identifier: Apache-2.0, MIT
 
+use super::Gas;
+
 /// Single gas charge in the VM. Contains information about what gas was for, as well
 /// as the amount of gas needed for computation and storage respectively.
 pub struct GasCharge<'a> {
     pub name: &'a str,
-    pub compute_gas: i64,
-    pub storage_gas: i64,
+    /// Compute costs
+    pub compute_gas: Gas,
+    /// Storage costs
+    pub storage_gas: Gas,
 }
 
 impl<'a> GasCharge<'a> {
-    pub fn new(name: &'a str, compute_gas: i64, storage_gas: i64) -> Self {
+    pub fn new(name: &'a str, compute_gas: Gas, storage_gas: Gas) -> Self {
         Self {
             name,
             compute_gas,
@@ -18,8 +22,9 @@ impl<'a> GasCharge<'a> {
         }
     }
 
-    /// Calculates total gas charge based on compute and storage multipliers.
-    pub fn total(&self) -> i64 {
+    /// Calculates total gas charge (in milligas) by summing compute and
+    /// storage gas associated with this charge.
+    pub fn total(&self) -> Gas {
         self.compute_gas + self.storage_gas
     }
 }
