@@ -17,43 +17,43 @@ use lazy_static::lazy_static;
 use num_traits::Zero;
 
 use super::GasCharge;
+use crate::gas::Gas;
 
 lazy_static! {
     static ref OH_SNAP_PRICES: PriceList = PriceList {
-        compute_gas_multiplier: 1,
         storage_gas_multiplier: 1300,
 
-        on_chain_message_compute_base: 38863,
-        on_chain_message_storage_base: 36,
-        on_chain_message_storage_per_byte: 1,
+        on_chain_message_compute_base: Gas::new(38863),
+        on_chain_message_storage_base: Gas::new(36),
+        on_chain_message_storage_per_byte: Gas::new(1),
 
-        on_chain_return_value_per_byte: 1,
+        on_chain_return_value_per_byte: Gas::new(1),
 
-        send_base: 29233,
-        send_transfer_funds: 27500,
-        send_transfer_only_premium: 159672,
-        send_invoke_method: -5377,
+        send_base: Gas::new(29233),
+        send_transfer_funds: Gas::new(27500),
+        send_transfer_only_premium: Gas::new(159672),
+        send_invoke_method: Gas::new(-5377),
 
-        create_actor_compute: 1108454,
-        create_actor_storage: 36 + 40,
-        delete_actor: -(36 + 40),
+        create_actor_compute: Gas::new(1108454),
+        create_actor_storage: Gas::new(36 + 40),
+        delete_actor: Gas::new(-(36 + 40)),
 
-        bls_sig_cost: 16598605,
-        secp256k1_sig_cost: 1637292,
+        bls_sig_cost: Gas::new(16598605),
+        secp256k1_sig_cost: Gas::new(1637292),
 
-        hashing_base: 31355,
-        compute_unsealed_sector_cid_base: 98647,
-        verify_seal_base: 2000, // TODO revisit potential removal of this
+        hashing_base: Gas::new(31355),
+        compute_unsealed_sector_cid_base: Gas::new(98647),
+        verify_seal_base: Gas::new(2000), // TODO revisit potential removal of this
 
-        verify_aggregate_seal_base: 0,
+        verify_aggregate_seal_base: Zero::zero(),
         verify_aggregate_seal_per: [
             (
                 RegisteredSealProof::StackedDRG32GiBV1P1,
-                449900
+                Gas::new(449900)
             ),
             (
                 RegisteredSealProof::StackedDRG64GiBV1P1,
-                359272
+                Gas::new(359272)
             )
         ].iter().copied().collect(),
         verify_aggregate_seal_steps: [
@@ -61,14 +61,14 @@ lazy_static! {
                 RegisteredSealProof::StackedDRG32GiBV1P1,
                 StepCost (
                     vec![
-                        Step{start: 4, cost: 103994170},
-                        Step{start: 7, cost: 112356810},
-                        Step{start: 13, cost: 122912610},
-                        Step{start: 26, cost: 137559930},
-                        Step{start: 52, cost: 162039100},
-                        Step{start: 103, cost: 210960780},
-                        Step{start: 205, cost: 318351180},
-                        Step{start: 410, cost: 528274980},
+                        Step{start: 4, cost: Gas::new(103994170)},
+                        Step{start: 7, cost: Gas::new(112356810)},
+                        Step{start: 13, cost: Gas::new(122912610)},
+                        Step{start: 26, cost: Gas::new(137559930)},
+                        Step{start: 52, cost: Gas::new(162039100)},
+                        Step{start: 103, cost: Gas::new(210960780)},
+                        Step{start: 205, cost: Gas::new(318351180)},
+                        Step{start: 410, cost: Gas::new(528274980)},
                     ]
                 )
             ),
@@ -76,14 +76,14 @@ lazy_static! {
                 RegisteredSealProof::StackedDRG64GiBV1P1,
                 StepCost (
                     vec![
-                        Step{start: 4, cost: 102581240},
-                        Step{start: 7, cost: 110803030},
-                        Step{start: 13, cost: 120803700},
-                        Step{start: 26, cost: 134642130},
-                        Step{start: 52, cost: 157357890},
-                        Step{start: 103, cost: 203017690},
-                        Step{start: 205, cost: 304253590},
-                        Step{start: 410, cost: 509880640},
+                        Step{start: 4, cost: Gas::new(102581240)},
+                        Step{start: 7, cost: Gas::new(110803030)},
+                        Step{start: 13, cost: Gas::new(120803700)},
+                        Step{start: 26, cost: Gas::new(134642130)},
+                        Step{start: 52, cost: Gas::new(157357890)},
+                        Step{start: 103, cost: Gas::new(203017690)},
+                        Step{start: 205, cost: Gas::new(304253590)},
+                        Step{start: 410, cost: Gas::new(509880640)},
                     ]
                 )
             )
@@ -91,28 +91,28 @@ lazy_static! {
         .cloned()
         .collect(),
 
-        verify_consensus_fault: 495422,
-        verify_replica_update: 36316136,
+        verify_consensus_fault: Gas::new(495422),
+        verify_replica_update: Gas::new(36316136),
         verify_post_lookup: [
             (
                 RegisteredPoStProof::StackedDRGWindow512MiBV1,
                 ScalingCost {
-                    flat: 117680921,
-                    scale: 43780,
+                    flat: Gas::new(117680921),
+                    scale: Gas::new(43780),
                 },
             ),
             (
                 RegisteredPoStProof::StackedDRGWindow32GiBV1,
                 ScalingCost {
-                    flat: 117680921,
-                    scale: 43780,
+                    flat: Gas::new(117680921),
+                    scale: Gas::new(43780),
                 },
             ),
             (
                 RegisteredPoStProof::StackedDRGWindow64GiBV1,
                 ScalingCost {
-                    flat: 117680921,
-                    scale: 43780,
+                    flat: Gas::new(117680921),
+                    scale: Gas::new(43780),
                 },
             ),
         ]
@@ -120,59 +120,65 @@ lazy_static! {
         .copied()
         .collect(),
 
-        get_randomness_base: 0,
-        get_randomness_per_byte: 0,
+        get_randomness_base: Zero::zero(),
+        get_randomness_per_byte: Zero::zero(),
 
-        block_memcpy_per_byte_cost: 0,
-        block_io_per_byte_cost: 0,
-        block_link_per_byte_cost: 1,
+        block_memcpy_per_byte_cost: Zero::zero(),
 
-        block_open_base: 114617,
-        block_read_base: 0,
-        block_create_base: 0,
-        block_link_base: 353640,
-        block_stat: 0,
+        block_open_base: Gas::new(114617),
+        block_open_memret_per_byte_cost: Zero::zero(),
+
+        block_link_base: Gas::new(353640),
+        block_link_storage_per_byte_cost: Gas::new(1),
+
+        block_create_base: Zero::zero(),
+        block_create_memret_per_byte_cost: Zero::zero(),
+
+        block_read_base: Zero::zero(),
+        block_stat_base: Zero::zero(),
+
+        syscall_cost: Zero::zero(),
+        extern_cost: Zero::zero(),
 
         wasm_rules: WasmGasPrices{
-            exec_instruction_cost_milli: 0,
+            exec_instruction_cost: Zero::zero(),
         },
     };
 
     static ref SKYR_PRICES: PriceList = PriceList {
-        compute_gas_multiplier: 1,
         storage_gas_multiplier: 1300,
 
-        on_chain_message_compute_base: 38863,
-        on_chain_message_storage_base: 36,
-        on_chain_message_storage_per_byte: 1,
+        on_chain_message_compute_base: Gas::new(38863),
+        on_chain_message_storage_base: Gas::new(36),
+        on_chain_message_storage_per_byte: Gas::new(1),
 
-        on_chain_return_value_per_byte: 1,
+        on_chain_return_value_per_byte: Gas::new(1),
 
-        send_base: 29233,
-        send_transfer_funds: 27500,
-        send_transfer_only_premium: 159672,
-        send_invoke_method: -5377,
+        send_base: Gas::new(29233),
+        send_transfer_funds: Gas::new(27500),
+        send_transfer_only_premium: Gas::new(159672),
+        send_invoke_method: Gas::new(-5377),
 
-        create_actor_compute: 1108454,
-        create_actor_storage: 36 + 40,
-        delete_actor: -(36 + 40),
+        create_actor_compute: Gas::new(1108454),
+        create_actor_storage: Gas::new(36 + 40),
+        delete_actor: Gas::new(-(36 + 40)),
 
-        bls_sig_cost: 16598605,
-        secp256k1_sig_cost: 1637292,
+        bls_sig_cost: Gas::new(16598605),
+        secp256k1_sig_cost: Gas::new(1637292),
 
-        hashing_base: 31355,
-        compute_unsealed_sector_cid_base: 98647,
-        verify_seal_base: 2000, // TODO revisit potential removal of this
+        hashing_base: Gas::new(31355),
+        compute_unsealed_sector_cid_base: Gas::new(98647),
+        verify_seal_base: Gas::new(2000), // TODO revisit potential removal of this
 
-        verify_aggregate_seal_base: 0,
+        verify_aggregate_seal_base: Zero::zero(),
         verify_aggregate_seal_per: [
             (
                 RegisteredSealProof::StackedDRG32GiBV1P1,
-                449900
+                Gas::new(449900)
             ),
             (
                 RegisteredSealProof::StackedDRG64GiBV1P1,
-                359272
+                Gas::new(359272)
             )
         ].iter().copied().collect(),
         verify_aggregate_seal_steps: [
@@ -180,14 +186,14 @@ lazy_static! {
                 RegisteredSealProof::StackedDRG32GiBV1P1,
                 StepCost (
                     vec![
-                        Step{start: 4, cost: 103994170},
-                        Step{start: 7, cost: 112356810},
-                        Step{start: 13, cost: 122912610},
-                        Step{start: 26, cost: 137559930},
-                        Step{start: 52, cost: 162039100},
-                        Step{start: 103, cost: 210960780},
-                        Step{start: 205, cost: 318351180},
-                        Step{start: 410, cost: 528274980},
+                        Step{start: 4, cost: Gas::new(103994170)},
+                        Step{start: 7, cost: Gas::new(112356810)},
+                        Step{start: 13, cost: Gas::new(122912610)},
+                        Step{start: 26, cost: Gas::new(137559930)},
+                        Step{start: 52, cost: Gas::new(162039100)},
+                        Step{start: 103, cost: Gas::new(210960780)},
+                        Step{start: 205, cost: Gas::new(318351180)},
+                        Step{start: 410, cost: Gas::new(528274980)},
                     ]
                 )
             ),
@@ -195,14 +201,14 @@ lazy_static! {
                 RegisteredSealProof::StackedDRG64GiBV1P1,
                 StepCost (
                     vec![
-                        Step{start: 4, cost: 102581240},
-                        Step{start: 7, cost: 110803030},
-                        Step{start: 13, cost: 120803700},
-                        Step{start: 26, cost: 134642130},
-                        Step{start: 52, cost: 157357890},
-                        Step{start: 103, cost: 203017690},
-                        Step{start: 205, cost: 304253590},
-                        Step{start: 410, cost: 509880640},
+                        Step{start: 4, cost: Gas::new(102581240)},
+                        Step{start: 7, cost: Gas::new(110803030)},
+                        Step{start: 13, cost: Gas::new(120803700)},
+                        Step{start: 26, cost: Gas::new(134642130)},
+                        Step{start: 52, cost: Gas::new(157357890)},
+                        Step{start: 103, cost: Gas::new(203017690)},
+                        Step{start: 205, cost: Gas::new(304253590)},
+                        Step{start: 410, cost: Gas::new(509880640)},
                     ]
                 )
             )
@@ -211,28 +217,28 @@ lazy_static! {
         .collect(),
 
         // TODO: PARAM_FINISH: this may need to be increased to account for the cost of an extern
-        verify_consensus_fault: 495422,
-        verify_replica_update: 36316136,
+        verify_consensus_fault: Gas::new(495422),
+        verify_replica_update: Gas::new(36316136),
         verify_post_lookup: [
             (
                 RegisteredPoStProof::StackedDRGWindow512MiBV1,
                 ScalingCost {
-                    flat: 117680921,
-                    scale: 43780,
+                    flat: Gas::new(117680921),
+                    scale: Gas::new(43780),
                 },
             ),
             (
                 RegisteredPoStProof::StackedDRGWindow32GiBV1,
                 ScalingCost {
-                    flat: 117680921,
-                    scale: 43780,
+                    flat: Gas::new(117680921),
+                    scale: Gas::new(43780),
                 },
             ),
             (
                 RegisteredPoStProof::StackedDRGWindow64GiBV1,
                 ScalingCost {
-                    flat: 117680921,
-                    scale: 43780,
+                    flat: Gas::new(117680921),
+                    scale: Gas::new(43780),
                 },
             ),
         ]
@@ -240,38 +246,36 @@ lazy_static! {
         .copied()
         .collect(),
 
-        block_memcpy_per_byte_cost: 4,
-        block_io_per_byte_cost: 2,
-        block_link_per_byte_cost: 1,
-        // TODO: PARAM_FINISH
+        get_randomness_base: Zero::zero(),
+        get_randomness_per_byte: Zero::zero(),
 
-        // TODO: PARAM_FINISH
-        get_randomness_base: 1,
-        // TODO: PARAM_FINISH
-        get_randomness_per_byte: 1,
+        block_memcpy_per_byte_cost: Gas::from_milligas(500),
 
-        // TODO: PARAM_FINIuiSH
-        block_open_base: 1,
-        // TODO: PARAM_FINISH
-        block_read_base: 1,
-        // TODO: PARAM_FINISH
-        block_create_base: 1,
-        // TODO: PARAM_FINISH
-        block_link_base: 1,
-        // TODO: PARAM_FINISH
-        block_stat: 1,
+        block_open_base: Gas::new(114617),
+        block_open_memret_per_byte_cost: Gas::new(10),
+
+        block_link_base: Gas::new(353640),
+        block_link_storage_per_byte_cost: Gas::new(1),
+
+        block_create_base: Zero::zero(),
+        block_create_memret_per_byte_cost: Gas::new(10),
+
+        block_read_base: Zero::zero(),
+        block_stat_base: Zero::zero(),
+
+        syscall_cost: Gas::new(14000),
+        extern_cost: Gas::new(21000),
 
         wasm_rules: WasmGasPrices{
-            exec_instruction_cost_milli: 5000,
+            exec_instruction_cost: Gas::new(4),
         },
-        // TODO: PARAM_FINISH
     };
 }
 
 #[derive(Clone, Debug, Copy)]
 pub(crate) struct ScalingCost {
-    flat: i64,
-    scale: i64,
+    flat: Gas,
+    scale: Gas,
 }
 
 #[derive(Clone, Debug)]
@@ -280,11 +284,11 @@ pub(crate) struct StepCost(Vec<Step>);
 #[derive(Clone, Debug, Copy)]
 pub(crate) struct Step {
     start: i64,
-    cost: i64,
+    cost: Gas,
 }
 
 impl StepCost {
-    pub(crate) fn lookup(&self, x: i64) -> i64 {
+    pub(crate) fn lookup(&self, x: i64) -> Gas {
         let mut i: i64 = 0;
         while i < self.0.len() as i64 {
             if self.0[i as usize].start > x {
@@ -294,20 +298,16 @@ impl StepCost {
         }
         i -= 1;
         if i < 0 {
-            return 0;
+            return Gas::zero();
         }
         self.0[i as usize].cost
     }
 }
 
-/// Provides prices for operations in the VM
+/// Provides prices for operations in the VM.
+/// All costs are in milligas.
 #[derive(Clone, Debug)]
 pub struct PriceList {
-    /// Compute gas charge multiplier
-    // * This multiplier is not currently applied to anything, but is matching lotus.
-    // * If the possible values are non 1 or if Lotus adds, we should change also.
-    #[allow(unused)]
-    pub(crate) compute_gas_multiplier: i64,
     /// Storage gas charge multiplier
     pub(crate) storage_gas_multiplier: i64,
 
@@ -317,83 +317,104 @@ pub struct PriceList {
     /// Together, these account for the cost of message propagation and validation,
     /// up to but excluding any actual processing by the VM.
     /// This is the cost a block producer burns when including an invalid message.
-    pub(crate) on_chain_message_compute_base: i64,
-    pub(crate) on_chain_message_storage_base: i64,
-    pub(crate) on_chain_message_storage_per_byte: i64,
+    pub(crate) on_chain_message_compute_base: Gas,
+    pub(crate) on_chain_message_storage_base: Gas,
+    pub(crate) on_chain_message_storage_per_byte: Gas,
 
     /// Gas cost charged to the originator of a non-nil return value produced
     /// by an on-chain message is given by:
     ///   len(return value)*OnChainReturnValuePerByte
-    pub(crate) on_chain_return_value_per_byte: i64,
+    pub(crate) on_chain_return_value_per_byte: Gas,
 
     /// Gas cost for any message send execution(including the top-level one
     /// initiated by an on-chain message).
     /// This accounts for the cost of loading sender and receiver actors and
     /// (for top-level messages) incrementing the sender's sequence number.
     /// Load and store of actor sub-state is charged separately.
-    pub(crate) send_base: i64,
+    pub(crate) send_base: Gas,
 
     /// Gas cost charged, in addition to SendBase, if a message send
     /// is accompanied by any nonzero currency amount.
     /// Accounts for writing receiver's new balance (the sender's state is
     /// already accounted for).
-    pub(crate) send_transfer_funds: i64,
+    pub(crate) send_transfer_funds: Gas,
 
     /// Gas cost charged, in addition to SendBase, if message only transfers funds.
-    pub(crate) send_transfer_only_premium: i64,
+    pub(crate) send_transfer_only_premium: Gas,
 
     /// Gas cost charged, in addition to SendBase, if a message invokes
     /// a method on the receiver.
     /// Accounts for the cost of loading receiver code and method dispatch.
-    pub(crate) send_invoke_method: i64,
+    pub(crate) send_invoke_method: Gas,
 
     /// Gas cost for creating a new actor (via InitActor's Exec method).
     /// Note: this costs assume that the extra will be partially or totally refunded while
     /// the base is covering for the put.
-    pub(crate) create_actor_compute: i64,
-    pub(crate) create_actor_storage: i64,
+    pub(crate) create_actor_compute: Gas,
+    pub(crate) create_actor_storage: Gas,
 
     /// Gas cost for deleting an actor.
     /// Note: this partially refunds the create cost to incentivise the deletion of the actors.
-    pub(crate) delete_actor: i64,
+    pub(crate) delete_actor: Gas,
 
     /// Gas cost for verifying bls signature
-    pub(crate) bls_sig_cost: i64,
+    pub(crate) bls_sig_cost: Gas,
     /// Gas cost for verifying secp256k1 signature
-    pub(crate) secp256k1_sig_cost: i64,
+    pub(crate) secp256k1_sig_cost: Gas,
 
-    pub(crate) hashing_base: i64,
+    pub(crate) hashing_base: Gas,
 
-    pub(crate) compute_unsealed_sector_cid_base: i64,
-    pub(crate) verify_seal_base: i64,
+    pub(crate) compute_unsealed_sector_cid_base: Gas,
+    pub(crate) verify_seal_base: Gas,
     #[allow(unused)]
-    pub(crate) verify_aggregate_seal_base: i64,
-    pub(crate) verify_aggregate_seal_per: AHashMap<RegisteredSealProof, i64>,
+    pub(crate) verify_aggregate_seal_base: Gas,
+    pub(crate) verify_aggregate_seal_per: AHashMap<RegisteredSealProof, Gas>,
     pub(crate) verify_aggregate_seal_steps: AHashMap<RegisteredSealProof, StepCost>,
 
     pub(crate) verify_post_lookup: AHashMap<RegisteredPoStProof, ScalingCost>,
-    pub(crate) verify_consensus_fault: i64,
-    pub(crate) verify_replica_update: i64,
+    pub(crate) verify_consensus_fault: Gas,
+    pub(crate) verify_replica_update: Gas,
 
-    pub(crate) get_randomness_base: i64,
-    pub(crate) get_randomness_per_byte: i64,
+    /// Gas cost for fetching randomness.
+    pub(crate) get_randomness_base: Gas,
+    /// Gas cost per every byte of randomness fetched.
+    pub(crate) get_randomness_per_byte: Gas,
 
-    pub(crate) block_memcpy_per_byte_cost: i64,
-    pub(crate) block_io_per_byte_cost: i64,
-    pub(crate) block_link_per_byte_cost: i64,
+    /// Gas cost per every block byte memcopied across boundaries.
+    pub(crate) block_memcpy_per_byte_cost: Gas,
 
-    pub(crate) block_open_base: i64,
-    pub(crate) block_read_base: i64,
-    pub(crate) block_create_base: i64,
-    pub(crate) block_link_base: i64,
-    pub(crate) block_stat: i64,
+    /// Gas cost for opening a block.
+    pub(crate) block_open_base: Gas,
+    /// Gas cost for every byte retained in FVM space when opening a block.
+    pub(crate) block_open_memret_per_byte_cost: Gas,
 
+    /// Gas cost for linking a block.
+    pub(crate) block_link_base: Gas,
+    /// Multiplier for storage gas per byte.
+    pub(crate) block_link_storage_per_byte_cost: Gas,
+
+    /// Gas cost for creating a block.
+    pub(crate) block_create_base: Gas,
+    /// Gas cost for every byte retained in FVM space when writing a block.
+    pub(crate) block_create_memret_per_byte_cost: Gas,
+
+    /// Gas cost for reading a block into actor space.
+    pub(crate) block_read_base: Gas,
+    /// Gas cost for statting a block.
+    pub(crate) block_stat_base: Gas,
+
+    /// General gas cost for performing a syscall, accounting for the overhead thereof.
+    pub(crate) syscall_cost: Gas,
+    /// General gas cost for calling an extern, accounting for the overhead thereof.
+    pub(crate) extern_cost: Gas,
+
+    /// Rules for execution gas.
     pub(crate) wasm_rules: WasmGasPrices,
 }
 
 #[derive(Clone, Debug, Eq, PartialEq, Hash)]
 pub struct WasmGasPrices {
-    pub(crate) exec_instruction_cost_milli: u64,
+    pub(crate) exec_instruction_cost: Gas,
 }
 
 impl PriceList {
@@ -408,15 +429,17 @@ impl PriceList {
                 * self.storage_gas_multiplier,
         )
     }
+
     /// Returns the gas required for storing the response of a message in the chain.
     #[inline]
     pub fn on_chain_return_value(&self, data_size: usize) -> GasCharge<'static> {
         GasCharge::new(
             "OnChainReturnValue",
-            0,
-            data_size as i64 * self.on_chain_return_value_per_byte * self.storage_gas_multiplier,
+            Zero::zero(),
+            self.on_chain_return_value_per_byte * data_size as i64 * self.storage_gas_multiplier,
         )
     }
+
     /// Returns the gas required when invoking a method.
     #[inline]
     pub fn on_method_invocation(
@@ -434,8 +457,14 @@ impl PriceList {
         if method_num != METHOD_SEND {
             ret += self.send_invoke_method;
         }
-        GasCharge::new("OnMethodInvocation", ret, 0)
+        GasCharge::new("OnMethodInvocation", ret, Zero::zero())
     }
+
+    /// Returns the gas cost to be applied on a syscall.
+    pub fn on_syscall(&self) -> GasCharge<'static> {
+        GasCharge::new("OnSyscall", self.syscall_cost, Zero::zero())
+    }
+
     /// Returns the gas required for creating an actor.
     #[inline]
     pub fn on_create_actor(&self) -> GasCharge<'static> {
@@ -445,15 +474,17 @@ impl PriceList {
             self.create_actor_storage * self.storage_gas_multiplier,
         )
     }
+
     /// Returns the gas required for deleting an actor.
     #[inline]
     pub fn on_delete_actor(&self) -> GasCharge<'static> {
         GasCharge::new(
             "OnDeleteActor",
-            0,
+            Zero::zero(),
             self.delete_actor * self.storage_gas_multiplier,
         )
     }
+
     /// Returns gas required for signature verification.
     #[inline]
     pub fn on_verify_signature(&self, sig_type: SignatureType) -> GasCharge<'static> {
@@ -461,13 +492,15 @@ impl PriceList {
             SignatureType::BLS => self.bls_sig_cost,
             SignatureType::Secp256k1 => self.secp256k1_sig_cost,
         };
-        GasCharge::new("OnVerifySignature", val, 0)
+        GasCharge::new("OnVerifySignature", val, Zero::zero())
     }
+
     /// Returns gas required for hashing data.
     #[inline]
     pub fn on_hashing(&self, _: usize) -> GasCharge<'static> {
-        GasCharge::new("OnHashing", self.hashing_base, 0)
+        GasCharge::new("OnHashing", self.hashing_base, Zero::zero())
     }
+
     /// Returns gas required for computing unsealed sector Cid.
     #[inline]
     pub fn on_compute_unsealed_sector_cid(
@@ -478,13 +511,14 @@ impl PriceList {
         GasCharge::new(
             "OnComputeUnsealedSectorCid",
             self.compute_unsealed_sector_cid_base,
-            0,
+            Zero::zero(),
         )
     }
+
     /// Returns gas required for seal verification.
     #[inline]
     pub fn on_verify_seal(&self, _info: &SealVerifyInfo) -> GasCharge<'static> {
-        GasCharge::new("OnVerifySeal", self.verify_seal_base, 0)
+        GasCharge::new("OnVerifySeal", self.verify_seal_base, Zero::zero())
     }
     #[inline]
     pub fn on_verify_aggregate_seals(
@@ -492,7 +526,7 @@ impl PriceList {
         aggregate: &AggregateSealVerifyProofAndInfos,
     ) -> GasCharge<'static> {
         let proof_type = aggregate.seal_proof;
-        let per_proof = self
+        let per_proof = *self
             .verify_aggregate_seal_per
             .get(&proof_type)
             .unwrap_or_else(|| {
@@ -518,14 +552,20 @@ impl PriceList {
         GasCharge::new(
             "OnVerifyAggregateSeals",
             per_proof * num + step.lookup(num),
-            0,
+            Zero::zero(),
         )
     }
+
     /// Returns gas required for replica verification.
     #[inline]
     pub fn on_verify_replica_update(&self, _replica: &ReplicaUpdateInfo) -> GasCharge<'static> {
-        GasCharge::new("OnVerifyReplicaUpdate", self.verify_replica_update, 0)
+        GasCharge::new(
+            "OnVerifyReplicaUpdate",
+            self.verify_replica_update,
+            Zero::zero(),
+        )
     }
+
     /// Returns gas required for PoSt verification.
     #[inline]
     pub fn on_verify_post(&self, info: &WindowPoStVerifyInfo) -> GasCharge<'static> {
@@ -540,14 +580,19 @@ impl PriceList {
                 .expect("512MiB lookup must exist in price table")
         });
 
-        let gas_used = cost.flat + info.challenged_sectors.len() as i64 * cost.scale;
+        let gas_used = cost.flat + cost.scale * info.challenged_sectors.len() as i64;
 
-        GasCharge::new("OnVerifyPost", gas_used, 0)
+        GasCharge::new("OnVerifyPost", gas_used, Zero::zero())
     }
+
     /// Returns gas required for verifying consensus fault.
     #[inline]
     pub fn on_verify_consensus_fault(&self) -> GasCharge<'static> {
-        GasCharge::new("OnVerifyConsensusFault", self.verify_consensus_fault, 0)
+        GasCharge::new(
+            "OnVerifyConsensusFault",
+            self.extern_cost + self.verify_consensus_fault,
+            Zero::zero(),
+        )
     }
 
     /// Returns the cost of the gas required for getting randomness from the client, based on the
@@ -556,92 +601,114 @@ impl PriceList {
     pub fn on_get_randomness(&self, entropy_size: usize) -> GasCharge<'static> {
         GasCharge::new(
             "OnGetRandomness",
-            self.get_randomness_base.saturating_add(
-                self.get_randomness_per_byte
-                    .saturating_mul(entropy_size as i64),
-            ),
-            0,
+            self.extern_cost
+                + self.get_randomness_base
+                + (self.get_randomness_per_byte * entropy_size as i64),
+            Zero::zero(),
         )
     }
 
     /// Returns the base gas required for loading an object, independent of the object's size.
     #[inline]
     pub fn on_block_open_base(&self) -> GasCharge<'static> {
-        GasCharge::new("OnBlockOpenBase", self.block_open_base, 0)
+        GasCharge::new(
+            "OnBlockOpenBase",
+            self.extern_cost + self.block_open_base,
+            Zero::zero(),
+        )
     }
 
     /// Returns the gas required for loading an object based on the size of the object.
     #[inline]
     pub fn on_block_open_per_byte(&self, data_size: usize) -> GasCharge<'static> {
-        // TODO: Should we also throw on a memcpy cost here (see https://github.com/filecoin-project/FIPs/blob/master/FIPS/fip-0032.md#ipld-state-management-fees)
+        let size = data_size as i64;
         GasCharge::new(
             "OnBlockOpenPerByte",
-            self.block_io_per_byte_cost.saturating_mul(data_size as i64),
-            0,
+            (self.block_open_memret_per_byte_cost * size)
+                + (self.block_memcpy_per_byte_cost * size),
+            Zero::zero(),
         )
     }
+
     /// Returns the gas required for reading a loaded object.
     #[inline]
     pub fn on_block_read(&self, data_size: usize) -> GasCharge<'static> {
         GasCharge::new(
             "OnBlockRead",
-            self.block_read_base.saturating_add(
-                self.block_memcpy_per_byte_cost
-                    .saturating_mul(data_size as i64),
-            ),
-            0,
+            self.block_read_base + (self.block_memcpy_per_byte_cost * data_size as i64),
+            Zero::zero(),
         )
     }
 
     /// Returns the gas required for adding an object to the FVM cache.
     #[inline]
     pub fn on_block_create(&self, data_size: usize) -> GasCharge<'static> {
+        let size = data_size as i64;
+        let mem_costs = (self.block_create_memret_per_byte_cost * size)
+            + (self.block_memcpy_per_byte_cost * size);
         GasCharge::new(
             "OnBlockCreate",
-            self.block_create_base.saturating_add(
-                self.block_memcpy_per_byte_cost
-                    .saturating_mul(data_size as i64),
-            ),
-            0,
+            self.block_create_base + mem_costs,
+            Zero::zero(),
         )
     }
 
     /// Returns the gas required for committing an object to the state blockstore.
     #[inline]
     pub fn on_block_link(&self, data_size: usize) -> GasCharge<'static> {
-        // TODO: The FIP makes it sound like this would need 2 memcpys, is that what's desired?
+        let size = data_size as i64;
+        let memcpy = self.block_memcpy_per_byte_cost * size;
         GasCharge::new(
             "OnBlockLink",
-            self.block_link_base,
-            // data_size as i64 * self.block_link_per_byte_cost * self.storage_gas_multiplier,
-            self.block_link_per_byte_cost
-                .saturating_mul(self.storage_gas_multiplier)
-                .saturating_mul(data_size as i64),
+            // twice the memcpy cost:
+            // - one from the block registry to the FVM BufferedBlockstore
+            // - one from the FVM BufferedBlockstore to the Node's Blockstore
+            //   when the machine finishes.
+            self.block_link_base + (memcpy * 2),
+            self.block_link_storage_per_byte_cost * self.storage_gas_multiplier * size,
         )
     }
 
     /// Returns the gas required for storing an object.
     #[inline]
     pub fn on_block_stat(&self) -> GasCharge<'static> {
-        GasCharge::new("OnBlockStat", self.block_stat, 0)
+        GasCharge::new("OnBlockStat", self.block_stat_base, Zero::zero())
     }
 }
 
 /// Returns gas price list by NetworkVersion for gas consumption.
 pub fn price_list_by_network_version(network_version: NetworkVersion) -> &'static PriceList {
     match network_version {
-        NetworkVersion::V14 | NetworkVersion::V15 => &OH_SNAP_PRICES,
+        NetworkVersion::V15 => &OH_SNAP_PRICES,
         _ => &SKYR_PRICES,
     }
 }
 
 impl Rules for WasmGasPrices {
-    fn instruction_cost(&self, _instruction: &Instruction) -> Option<u64> {
-        Some(self.exec_instruction_cost_milli)
+    fn instruction_cost(&self, instruction: &Instruction) -> Option<u64> {
+        if self.exec_instruction_cost.is_zero() {
+            return Some(0);
+        }
+
+        // Rules valid for nv16. We will need to be generic over Rules (massive
+        // generics tax), use &dyn Rules (which breaks other things), or pass
+        // in the network version, or rules version, to vary these prices going
+        // forward.
+        match instruction {
+            // FIP-0032: nop, drop, block, loop, unreachable, return, else, end are priced 0.
+            Instruction::Nop
+            | Instruction::Drop
+            | Instruction::Block(_)
+            | Instruction::Loop(_)
+            | Instruction::Unreachable
+            | Instruction::Return
+            | Instruction::Else
+            | Instruction::End => Some(0),
+            _ => Some(self.exec_instruction_cost.as_milligas() as u64),
+        }
     }
 
     fn memory_grow_cost(&self) -> MemoryGrowCost {
-        // todo use pricelist
         MemoryGrowCost::Free
     }
 }
