@@ -7,7 +7,7 @@
 //! of your choice during the initialization of the consuming application.
 
 pub use kernel::default::DefaultKernel;
-pub use kernel::{BlockError, Kernel};
+pub use kernel::Kernel;
 
 pub mod call_manager;
 pub mod executor;
@@ -37,10 +37,6 @@ pub mod init_actor;
 #[cfg(feature = "testing")]
 pub mod system_actor;
 
-mod market_actor;
-mod power_actor;
-mod reward_actor;
-
 pub mod trace;
 
 use cid::multihash::{Code, MultihashDigest};
@@ -76,7 +72,7 @@ mod test {
     impl Rand for DummyExterns {
         fn get_chain_randomness(
             &self,
-            _pers: fvm_shared::crypto::randomness::DomainSeparationTag,
+            _pers: i64,
             _round: fvm_shared::clock::ChainEpoch,
             _entropy: &[u8],
         ) -> anyhow::Result<[u8; 32]> {
@@ -85,7 +81,7 @@ mod test {
 
         fn get_beacon_randomness(
             &self,
-            _pers: fvm_shared::crypto::randomness::DomainSeparationTag,
+            _pers: i64,
             _round: fvm_shared::clock::ChainEpoch,
             _entropy: &[u8],
         ) -> anyhow::Result<[u8; 32]> {
@@ -119,7 +115,7 @@ mod test {
 
         let actors_cid = bs.put_cbor(&(0, manifest_cid), Code::Blake2b256).unwrap();
 
-        let mc = NetworkConfig::new(fvm_shared::version::NetworkVersion::V14)
+        let mc = NetworkConfig::new(fvm_shared::version::NetworkVersion::V15)
             .override_actors(actors_cid)
             .for_epoch(0, root);
 
