@@ -95,6 +95,15 @@ fn memory_and_data<'a, K: Kernel>(
     (Memory::new(mem), data)
 }
 
+macro_rules! charge_syscall_gas {
+    ($kernel:expr) => {
+        let charge = $kernel.price_list().on_syscall();
+        $kernel
+            .charge_gas(charge.name, charge.compute_gas)
+            .map_err(Abort::from_error_as_fatal)?;
+    };
+}
+
 // Unfortunately, we can't implement this for _all_ functions. So we implement it for functions of up to 6 arguments.
 macro_rules! impl_bind_syscalls {
     ($($t:ident)*) => {
@@ -118,6 +127,8 @@ macro_rules! impl_bind_syscalls {
                         charge_for_exec(&mut caller)?;
 
                         let (mut memory, mut data) = memory_and_data(&mut caller);
+                        charge_syscall_gas!(data.kernel);
+
                         let ctx = Context{kernel: &mut data.kernel, memory: &mut memory};
                         let out = syscall(ctx $(, $t)*).into();
 
@@ -130,7 +141,7 @@ macro_rules! impl_bind_syscalls {
                             Ok(Err(err)) => {
                                 let code = err.1;
                                 log::trace!("syscall {}::{}: fail ({})", module, name, code as u32);
-                                data.last_error = Some(backtrace::Cause::new(module, name, err));
+                                data.last_error = Some(backtrace::Cause::from_syscall(module, name, err));
                                 Ok(code as u32)
                             },
                             Err(e) => Err(e.into()),
@@ -146,12 +157,13 @@ macro_rules! impl_bind_syscalls {
                         charge_for_exec(&mut caller)?;
 
                         let (mut memory, mut data) = memory_and_data(&mut caller);
+                        charge_syscall_gas!(data.kernel);
 
                         // We need to check to make sure we can store the return value _before_ we do anything.
                         if (ret as u64) > (memory.len() as u64)
                             || memory.len() - (ret as usize) < mem::size_of::<Ret::Value>() {
                             let code = ErrorNumber::IllegalArgument;
-                            data.last_error = Some(backtrace::Cause::new(module, name, SyscallError(format!("no space for return value"), code)));
+                            data.last_error = Some(backtrace::Cause::from_syscall(module, name, SyscallError(format!("no space for return value"), code)));
                             return Ok(code as u32);
                         }
 
@@ -166,7 +178,7 @@ macro_rules! impl_bind_syscalls {
                             Ok(Err(err)) => {
                                 let code = err.1;
                                 log::trace!("syscall {}::{}: fail ({})", module, name, code as u32);
-                                data.last_error = Some(backtrace::Cause::new(module, name, err));
+                                data.last_error = Some(backtrace::Cause::from_syscall(module, name, err));
                                 Ok(code as u32)
                             },
                             Err(e) => Err(e.into()),
@@ -189,3 +201,4 @@ impl_bind_syscalls!(A B C);
 impl_bind_syscalls!(A B C D);
 impl_bind_syscalls!(A B C D E);
 impl_bind_syscalls!(A B C D E F);
+impl_bind_syscalls!(A B C D E F G);
