@@ -1,6 +1,7 @@
 use std::str;
 
 use super::Context;
+use crate::gas::Gas;
 use crate::kernel::{ClassifyResult, Result};
 use crate::Kernel;
 
@@ -12,5 +13,6 @@ pub fn charge_gas(
 ) -> Result<()> {
     let name =
         str::from_utf8(context.memory.try_slice(name_off, name_len)?).or_illegal_argument()?;
-    context.kernel.charge_gas(name, compute)
+    // Gas charges from actors are always in full gas units. We use milligas internally, so convert here.
+    context.kernel.charge_gas(name, Gas::new(compute))
 }
