@@ -1,10 +1,7 @@
-use anyhow::Context as _;
-use fvm_shared::crypto::randomness::DomainSeparationTag;
 use fvm_shared::randomness::RANDOMNESS_LENGTH;
-use num_traits::FromPrimitive;
 
 use super::Context;
-use crate::kernel::{ClassifyResult, Result};
+use crate::kernel::Result;
 use crate::Kernel;
 
 /// Gets 32 bytes of randomness from the ticket chain.
@@ -19,10 +16,6 @@ pub fn get_chain_randomness(
     entropy_len: u32,
 ) -> Result<[u8; RANDOMNESS_LENGTH]> {
     let entropy = context.memory.try_slice(entropy_off, entropy_len)?;
-    // TODO determine if this error should lead to an abort.
-    let pers = DomainSeparationTag::from_i64(pers)
-        .context("invalid domain separation tag")
-        .or_illegal_argument()?;
     context
         .kernel
         .get_randomness_from_tickets(pers, round, entropy)
@@ -40,10 +33,6 @@ pub fn get_beacon_randomness(
     entropy_len: u32,
 ) -> Result<[u8; RANDOMNESS_LENGTH]> {
     let entropy = context.memory.try_slice(entropy_off, entropy_len)?;
-    // TODO determine if this error should lead to an abort.
-    let pers = DomainSeparationTag::from_i64(pers)
-        .context("invalid domain separation tag")
-        .or_illegal_argument()?;
     context
         .kernel
         .get_randomness_from_beacon(pers, round, entropy)
