@@ -1,11 +1,11 @@
 use fvm_shared::error::ExitCode;
+use fvm_shared::sys::out::vm::InvocationContext;
 use fvm_shared::sys::SyscallSafe;
 use fvm_shared::version::NetworkVersion;
 
 use super::error::Abort;
 use super::Context;
-use crate::kernel::{ClassifyResult, Context as _};
-use crate::Kernel;
+use crate::kernel::{ClassifyResult, Kernel};
 
 /// An uninhabited type. We use this in `abort` to make sure there's no way to return without
 /// returning an error.
@@ -21,6 +21,8 @@ pub fn abort(
     message_off: u32,
     message_len: u32,
 ) -> Result<Never, Abort> {
+    use crate::kernel::Context as _;
+
     let code = ExitCode::new(code);
     if context.kernel.network_version() >= NetworkVersion::V16 && code.is_system_error() {
         return Err(Abort::Exit(
@@ -45,3 +47,21 @@ pub fn abort(
     };
     Err(Abort::Exit(code, message))
 }
+
+pub fn context(context: Context<'_, impl Kernel>) -> crate::kernel::Result<InvocationContext> {
+    use anyhow::Context as _;
+
+    Ok(InvocationContext {
+        caller: context.kernel.msg_caller(),
+        receiver: context.kernel.msg_receiver(),
+        method_number: context.kernel.msg_method_number(),
+        value_received: context
+            .kernel
+            .msg_value_received()
+            .try_into()
+            .context("invalid token amount")
+            .or_fatal()?,
+        network_curr_epoch: context.kernel.network_epoch(),
+        network_version: context.kernel.network_version() as u32,
+    })
+}
